#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <pwd.h>
#include <rsbac/types.h>
#include <rsbac/aci_data_structures.h>
#include <rsbac/getname.h>
#include <rsbac/acl_getname.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

int verbose=0;
int recurse=0;
int printall=0;
int effective=1;
union rsbac_attribute_value_t value;
enum rsbac_target_t target;
enum rsbac_attribute_t attr;
char * progname;
enum rsbac_acl_subject_type_t subj_type=ACLS_USER;
rsbac_acl_subject_id_t subj_id=RSBAC_NO_USER;

int process(char * name)
  {
    int res = 0;
    char tmp1[120];
    enum rsbac_target_t type=T_NONE;
    char type_n[5] = "";
    struct stat buf;
    struct rsbac_acl_syscall_arg_t   arg;
    struct rsbac_acl_syscall_n_arg_t arg_n;
    rsbac_acl_rights_vector_t rights_vector = 0;

    if(!strcmp(name,":DEFAULT:"))
      {
        if(target == T_FD)
          type = T_FILE;
        else
          type = target;
        switch(type)
          {
            case T_FILE:
              strcpy(type_n,"FILE");
              break;
            case T_DIR:
              strcpy(type_n,"DIR");
              break;
            case T_DEV:
              strcpy(type_n,"DEV");
              break;
            case T_IPC:
              strcpy(type_n,"IPC");
              arg.tid.ipc.type = I_none;
              break;
            case T_SCD:
              strcpy(type_n,"SCD");
              arg.tid.scd = AST_none;
              break;
            case T_PROCESS:
              strcpy(type_n,"PROCESS");
              arg.tid.process = 0;
              break;
            default:
              fprintf(stderr, "Invalid target %u!\n",
                      target);
              return(1);
          }
        arg_n.name = NULL;
      }
    else
      {
        switch(target)
          {
            case T_FILE:
            case T_DIR:
            case T_FD:
            case T_DEV:
              if (lstat(name,&buf))
                {
                  fprintf(stderr, "stat for file %s returned error: %s\n",
                          name,
                          strerror(errno));
                  return(1);
                }
            default:
              break;
          }
        arg_n.name = name;
        switch(target)
          {
            case T_FILE:
              if(   (S_ISREG(buf.st_mode))
                 || (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_FILE;
                  strcpy(type_n,"FILE");
                }
              else
                {
                  if(verbose)
                    printf("%s is no file, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_DIR:
              if(S_ISDIR(buf.st_mode))
                {
                  type = T_DIR;
                  strcpy(type_n,"DIR");
                }
              else
                {
                  if(verbose)
                    printf("%s is no dir, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_FD:
              if(S_ISDIR(buf.st_mode))
                {
                  type = T_DIR;
                  strcpy(type_n,"DIR");
                }
              else
              if(   (S_ISREG(buf.st_mode))
                 || (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_FILE;
                  strcpy(type_n,"FILE");
                }
              else
                {
                  if(verbose)
                    printf("%s is neither file nor dir, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_DEV:
              if(   (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_DEV;
                  strcpy(type_n,"DEV");
                }
              else
                {
                  if(verbose)
                    printf("%s is no file, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_SCD:
              arg.tid.scd = get_acl_scd_type_nr(name);
              if((arg.tid.scd == ST_none) || (arg.tid.scd == AST_none))
                {
                  fprintf(stderr, "%s is no valid SCD name, skipped\n",
                           name);
                  return(1);
                }
              break;

            default:
              fprintf(stderr, "Invalid target %u for %s, skipped!\n",
                      target, name);
              return(1);
          }
      }
    if(verbose)
    printf("Processing %s '%s'\n",
           type_n,
           name);
    switch(target)
      {
        case T_FILE:
        case T_DIR:
        case T_FD:
        case T_DEV:
          arg_n.target = type;
          arg_n.subj_type = subj_type;
          arg_n.subj_id = subj_id;
          res = rsbac_acl_get_rights_n(&arg_n, &rights_vector, effective);
          break;

        default:
          arg.target = target;
          arg.subj_type = subj_type;
          arg.subj_id = subj_id;
          res = rsbac_acl_get_rights(&arg, &rights_vector, effective);
      }
    if(res)
      {
        if((res > -RSBAC_EPERM) || (res < -RSBAC_EDECISIONMISMATCH))
          fprintf(stderr, "%s: error: %s\n", name, strerror(-res));
        else
          {
            get_error_name(tmp1,res);
            fprintf(stderr, "%s: error: %s\n", name, tmp1);
          }
      }
    else
      {
        printf("%-19s: %s\n",
               name, u64tostracl(tmp1, rights_vector));
        if(printall)
          {
            int i;

            for (i=0; i<R_NONE; i++)
              if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                printf("  %s\n", get_request_name(tmp1,i));
            for (i=RSBAC_ACL_SPECIAL_RIGHT_BASE; i<ACLR_NONE; i++)
              if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                printf("  %s\n", get_acl_special_right_name(tmp1,i));
          }
      }

do_recurse:
    if (   arg_n.name
        && (type == T_DIR)
        && recurse)
      {
        DIR * dir_stream_p;
        struct dirent * dirent_p;
        char name2[512];

        if(S_ISLNK(buf.st_mode))
          return(0);
        if(!(dir_stream_p = opendir(name)))
          {
            fprintf(stderr, "opendir for dir %s returned error: %s\n",
                   name,
                   strerror(errno));
            return(-2);
          }
        while((dirent_p = readdir(dir_stream_p)))
          {
            if(   (strcmp(".",dirent_p->d_name))
               && (strcmp("..",dirent_p->d_name)) )
              {
                strcpy(name2,name);
                strcat(name2,"/");
                strcat(name2,dirent_p->d_name);
                process(name2);
              }
          }
        closedir(dir_stream_p);
      }
    return(0);
  }

int main(int argc, char ** argv)
{
  int res = 0;
  int i;
  struct passwd * user_info_p;
  
  progname = argv[0];
  while((argc > 1) && (argv[1][0] == '-'))
    {
      char * pos = argv[1];
      pos++;
      while(*pos)
        {
          switch(*pos)
            {
              case 'v':
                verbose=1;
                break;
              case 'r':
                recurse=1;
                break;
              case 'p':
                printall=1;
                break;
              case 'd':
                effective=0;
                break;
              case 'n':
                {
                  char tmp[80];
                  for(i=0; i<ST_none; i++)
                    printf("%s\n", get_scd_type_name(tmp, i));
                  for(i=AST_min; i<AST_none; i++)
                    printf("%s\n", get_acl_scd_type_name(tmp, i));
                  exit(0);
                }
              case 'u':
                if(argc < 2)
                  {
                    fprintf(stderr, "%s: no user for switch u\n", progname);
                    exit(1);
                  }
                if(   (!(user_info_p = getpwnam(argv[2])))
                   && (   (!(user_info_p = getpwuid(strtol(argv[2],0,0))))
                   || (!(user_info_p->pw_uid) && strcmp("0",argv[2])) )
                  )
                  {
                    fprintf(stderr, "Invalid user %s!\n", argv[2]);
                    exit(1);
                  }
                subj_type = ACLS_USER;
                subj_id = user_info_p->pw_uid;
                printf("%s: User %s\n", progname, user_info_p->pw_name);
                argv++;
                argc--;
                break;
              case 'g':
                if(argc < 2)
                  {
                    fprintf(stderr, "%s: no group for switch g\n", progname);
                    exit(1);
                  }
                subj_type = ACLS_GROUP;
                subj_id = strtol(argv[2],0,0);
                printf("%s: Group %u\n", progname, subj_id);
                argv++;
                argc--;
                break;
              case 'l':
                if(argc < 2)
                  {
                    fprintf(stderr, "%s: no group for switch g\n", progname);
                    exit(1);
                  }
                subj_type = ACLS_ROLE;
                subj_id = strtol(argv[2],0,0);
                printf("%s: Role %u\n", progname, subj_id);
                argv++;
                argc--;
                break;
              default:
                fprintf(stderr, "%s: unknown parameter %c\n", progname, *pos);
            }
          pos++;
        }
      argv++;
      argc--;
    }
  if (argc > 1)
    {
      target = get_target_nr(argv[1]);
      if(target == T_NONE)
        {
          if(verbose)
            printf("%s: %i targets\n\n", progname, argc - 1);
          fprintf(stderr, "%s: No target type given, assuming FD\n", progname);
          target = T_FD;
          if(argc < 1)
            process(".");
          else
            for (i=1;i < (argc);i++)
              {
                process(argv[i]);
              }
        }
      else
        {
          if(argc > 2)
            {
              if(verbose)
                printf("%s: %i targets\n\n", progname, argc - 2);
              for (i=1;i < (argc-1);i++)
                {
                  process(argv[i+1]);
                }
            }
          else
            process(".");
        }
    }
  else
    {
      printf("%s (RSBAC %s)\n***\n", progname, VERSION);
      printf("Use: %s [switches] target-type file/dirname(s)\n", progname);  
      printf("- -v = verbose, -r = recurse into subdirs,\n");
      printf("- -p = print right names, -d = give direct, not effective rights\n");
      printf("- -n = list valid SCD names\n");
      printf("- -u user = print rights for given user, not caller\n");
      printf("- -g group = print rights for given group, not caller\n");
      printf("- -l role = print rights for given role, not caller\n");
      printf("- target-type = FILE, DIR, DEV or FD\n");
      printf("  (FD: let %s decide between FILE and DIR, no DEV),\n", progname);
      printf("  (IPC, PROCESS: only :DEFAULT:\n");
      printf("- Use name :DEFAULT: for default ACL\n");
    }
  return (res);
}

