#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <pwd.h>
#include <rsbac/types.h>
#include <rsbac/aci_data_structures.h>
#include <rsbac/getname.h>
#include <rsbac/acl_getname.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

#define NR_ENTRIES 20
#define GRANT_PROG "acl_grant"

int verbose=0;
int recurse=0;
int printall=0;
int backup=0;
union rsbac_attribute_value_t value;
enum rsbac_target_t target;
enum rsbac_attribute_t attr;
char * progname;

int process(char * name)
  {
    int res = 0;
    char tmp1[120], tmp2[120];
    enum rsbac_target_t type=T_NONE;
    char type_n[8] = "";
    struct stat buf;
    char * i_name;
    struct rsbac_acl_entry_t entry_array[NR_ENTRIES];
    union rsbac_target_id_t tid;

    if(!strcmp(name,":DEFAULT:"))
      {
        if(target == T_FD)
          type = T_FILE;
        else
          type = target;
        switch(type)
          {
            case T_FILE:
              strcpy(type_n,"FILE");
              break;
            case T_DIR:
              strcpy(type_n,"DIR");
              break;
            case T_DEV:
              strcpy(type_n,"DEV");
              break;
            case T_IPC:
              strcpy(type_n,"IPC");
              tid.ipc.type = I_none;
              break;
            case T_SCD:
              strcpy(type_n,"SCD");
              tid.scd = AST_none;
              break;
            case T_PROCESS:
              strcpy(type_n,"PROCESS");
              tid.process = 0;
              break;
            default:
              fprintf(stderr, "Invalid target %u!\n",
                      target);
              return(1);
          }
        i_name = NULL;
      }
    else
      {
        switch(target)
          {
            case T_FILE:
            case T_DIR:
            case T_FD:
            case T_DEV:
              if (lstat(name,&buf))
                {
                  fprintf(stderr, "stat for file %s returned error: %s\n",
                          name,
                          strerror(errno));
                  return(1);
                }
            default:
              break;
          }
        i_name = name;
        switch(target)
          {
            case T_FILE:
              if(   (S_ISREG(buf.st_mode))
                 || (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_FILE;
                  strcpy(type_n,"FILE");
                }
              else
                {
                  if(verbose)
                    printf("%s is no file, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_DIR:
              if(S_ISDIR(buf.st_mode))
                {
                  type = T_DIR;
                  strcpy(type_n,"DIR");
                }
              else
                {
                  if(verbose)
                    printf("%s is no dir, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_FD:
              if(S_ISDIR(buf.st_mode))
                {
                  type = T_DIR;
                  strcpy(type_n,"DIR");
                }
              else
              if(   (S_ISREG(buf.st_mode))
                 || (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_FILE;
                  strcpy(type_n,"FILE");
                }
              else
                {
                  if(verbose)
                    printf("%s is neither file nor dir, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_DEV:
              if(   (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_DEV;
                  strcpy(type_n,"DEV");
                }
              else
                {
                  if(verbose)
                    printf("%s is no file, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;

            case T_SCD:
              tid.scd = get_acl_scd_type_nr(name);
              if((tid.scd == ST_none) || (tid.scd == AST_none))
                {
                  fprintf(stderr, "%s is no valid SCD name, skipped\n",
                           name);
                  return(1);
                }
              break;

            default:
              fprintf(stderr, "Invalid target %u for %s, skipped!\n",
                      target, name);
              return(1);
          }
      }
    if(verbose)
      printf("Processing %s '%s'\n",
             type_n,
             name);

    switch(target)
      {
        case T_FILE:
        case T_DIR:
        case T_FD:
        case T_DEV:
          res = rsbac_acl_get_tlist_n(type, i_name, entry_array, NR_ENTRIES);
          break;
        default:
          res = rsbac_acl_get_tlist(target, &tid, entry_array, NR_ENTRIES);
      }
    if(res<0)
      {
        if((res > -RSBAC_EPERM) || (res < -RSBAC_EDECISIONMISMATCH))
          fprintf(stderr, "%s: error: %s\n", name, strerror(-res));
        else
          {
            get_error_name(tmp1,res);
            fprintf(stderr, "%s: error: %s\n", name, tmp1);
          }
      }
    else
      {
        int j;

        if(backup)
          {
            if(printall)
              {
                int i;

                for(j=0; j<res; j++)
                  {
                    printf("%s -vs %s %u",
                           GRANT_PROG,
                           get_acl_subject_type_name(tmp1, entry_array[j].subj_type),
                           entry_array[j].subj_id);
                    for (i=0; i<R_NONE; i++)
                      if(entry_array[j].rights & ((rsbac_acl_rights_vector_t) 1 << i))
                        printf(" %s", get_request_name(tmp1,i));
                    for (i=RSBAC_ACL_SPECIAL_RIGHT_BASE; i<ACLR_NONE; i++)
                      if(entry_array[j].rights & ((rsbac_acl_rights_vector_t) 1 << i))
                        printf(" %s", get_acl_special_right_name(tmp1,i));
                    printf(" %s %s\n", type_n, name);
                  }
              }
            else
              {
                for(j=0; j<res; j++)
                  {
                    printf("%s -vsb %s %u %s %s %s\n",
                           GRANT_PROG,
                           get_acl_subject_type_name(tmp1, entry_array[j].subj_type),
                           entry_array[j].subj_id,
                           u64tostracl(tmp2, entry_array[j].rights),
                           type_n,
                           name);
                  }
              }
          }
        else /* no backup */
          {
            printf("%s: %i entries\n",
                   name, res);
            for(j=0; j<res; j++)
              {
                printf("  %s %u:\t%s\n",
                       get_acl_subject_type_name(tmp1, entry_array[j].subj_type),
                       entry_array[j].subj_id,
                       u64tostracl(tmp2, entry_array[j].rights));
                if(printall)
                  {
                    int i;

                    for (i=0; i<R_NONE; i++)
                      if(entry_array[j].rights & ((rsbac_acl_rights_vector_t) 1 << i))
                        printf("    %s\n", get_request_name(tmp1,i));
                    for (i=RSBAC_ACL_SPECIAL_RIGHT_BASE; i<ACLR_NONE; i++)
                      if(entry_array[j].rights & ((rsbac_acl_rights_vector_t) 1 << i))
                        printf("    %s\n", get_acl_special_right_name(tmp1,i));
                  }
              }
          }
      }

do_recurse:
    if (   (type == T_DIR)
        && recurse)
      {
        DIR * dir_stream_p;
        struct dirent * dirent_p;
        char name2[512];

        if(S_ISLNK(buf.st_mode))
          return(0);
        if(!(dir_stream_p = opendir(name)))
          {
            fprintf(stderr, "opendir for dir %s returned error: %s\n",
                   name,
                   strerror(errno));
            return(-2);
          }
        while((dirent_p = readdir(dir_stream_p)))
          {
            if(   (strcmp(".",dirent_p->d_name))
               && (strcmp("..",dirent_p->d_name)) )
              {
                strcpy(name2,name);
                strcat(name2,"/");
                strcat(name2,dirent_p->d_name);
                process(name2);
              }
          }
        closedir(dir_stream_p);
      }
    return(0);
  }

int main(int argc, char ** argv)
{
  int res = 0;
  int i;
  
  progname = argv[0];
  while((argc > 1) && (argv[1][0] == '-'))
    {
      char * pos = argv[1];
      pos++;
      while(*pos)
        {
          switch(*pos)
            {
              case 'v':
                verbose=1;
                break;
              case 'r':
                recurse=1;
                break;
              case 'p':
                printall=1;
                break;
              case 'b':
                backup=1;
                break;
              case 'n':
                {
                  char tmp[80];
                  for(i=0; i<ST_none; i++)
                    printf("%s\n", get_scd_type_name(tmp, i));
                  for(i=AST_min; i<AST_none; i++)
                    printf("%s\n", get_acl_scd_type_name(tmp, i));
                  exit(0);
                }
              default:
                fprintf(stderr, "%s: unknown parameter %c\n", progname, *pos);
            }
          pos++;
        }
      argv++;
      argc--;
    }
  if (argc > 1)
    {
      target = get_target_nr(argv[1]);
      if(target == T_NONE)
        {
          if(verbose)
            printf("%s: %i targets\n\n", progname, argc - 1);
          fprintf(stderr, "%s: No target type given, assuming FD\n", progname);
          target = T_FD;
          if(argc < 1)
            process(".");
          else
            for (i=1;i < (argc);i++)
              {
                process(argv[i]);
              }
        }
      else
        {
          if(argc > 2)
            {
              if(verbose)
                printf("%s: %i targets\n\n", progname, argc - 2);
              for (i=1;i < (argc-1);i++)
                {
                  process(argv[i+1]);
                }
            }
          else
            process(".");
        }
    }
  else
    {
      printf("%s (RSBAC %s)\n***\n", progname, VERSION);
      printf("Use: %s [switches] target-type file/dir/scdname(s)\n", progname);  
      printf("- -v = verbose, -r = recurse into subdirs,\n");
      printf("- -p = print right names, -b = backup mode\n");
      printf("- -n = list valid SCD names\n");
      printf("- target-type = FILE, DIR, DEV, IPC, SCD, PROCESS or FD\n");
      printf("  (FD: let %s decide between FILE and DIR, no DEV),\n", progname);
      printf("  (IPC, PROCESS: only :DEFAULT:\n");
      printf("- Use name :DEFAULT: for default ACL\n");
    }
  return (res);
}

