#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <rsbac/types.h>
#include <rsbac/aci_data_structures.h>
#include <rsbac/getname.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

int main(int argc, char ** argv)
{
  int attr_list[RSBAC_USER_NR_ATTRIBUTES] = RSBAC_USER_ATTR_LIST;
  int res = 0;
  char tmp1[80],tmp2[120];
  int j;
  union rsbac_attribute_value_t value;
  union rsbac_target_id_t tid;
  enum rsbac_attribute_t attr;
  struct passwd * user_info_p;
  u_int position, catval;
  __u64 k;

  switch(argc)
    {
      case 4:
        attr = get_attribute_nr(argv[2]);
        if(attr == A_none)
          {
            fprintf(stderr, "%s: Invalid Attribute %s!\n", argv[0], argv[2]);
            exit(1);
          }
        if(   (!(user_info_p = getpwnam(argv[1])))
           && (   (!(user_info_p = getpwuid(strtol(argv[1],0,0))))
               || (!(user_info_p->pw_uid) && strcmp("0",argv[1])) )
          )
            {
              fprintf(stderr, "%s: Invalid user %s!\n\n", argv[0], argv[1]);
              exit(1);
            }
        tid.user = user_info_p->pw_uid;

        switch(attr)
          {
            case A_none:
              fprintf(stderr, "%s: Invalid attribute %s\n", argv[0], argv[3]);
              exit(1);
            case A_mac_categories:
              if(strlen(argv[3]) != RSBAC_MAC_NR_CATS)
                {
                  fprintf(stderr, "%s: Invalid attribute value, length must be %i\n", argv[0],
                          RSBAC_MAC_NR_CATS);
                  exit(1);
                }
              for(j=0;j<strlen(argv[4]);j++)
                {
                  if(   argv[3][j] != '0'
                     && argv[3][j] != '1')
                    {
                      fprintf(stderr, "%s: Invalid attribute value char, must be 0 or 1\n", argv[0]);
                      exit(1);
                    }
                }
              strtou64mac(argv[3], &value.mac_categories);
              break;

              default:
                value.dummy = strtol(argv[3],0,0);
          }
        res = rsbac_set_attr(T_USER, &tid, attr, &value);
        error_exit(res);
        exit(0);

      case 5:
        if(!strcmp(argv[2],"mac_categories"))
          {
            if(   (!(user_info_p = getpwnam(argv[1])))
               && (   (!(user_info_p = getpwuid(strtol(argv[1],0,0))))
                   || (!(user_info_p->pw_uid) && strcmp("0",argv[1])) )
              )
              {
                fprintf(stderr, "%s: Invalid user %s!\n\n", argv[0], argv[1]);
                exit(1);
              }
            tid.user = user_info_p->pw_uid;
            position = strtol(argv[3],0,10);
            if(position > RSBAC_MAC_MAX_CAT)
              {
                fprintf(stderr, "Invalid position counter %s\n", argv[4]);
                exit(1);
              }
            catval = strtol(argv[4],0,0);
            if(catval > 1)
              {
                fprintf(stderr, "Invalid value %s\n", argv[4]);
                exit(1);
              }
            res = rsbac_get_attr(T_USER, &tid, A_mac_categories, &value, FALSE);
            error_exit(res);
            k = ((__u64) 1) << position;
            if(catval)
              value.mac_categories |= k;
            else
              value.mac_categories &= ~k;
            res = rsbac_set_attr(T_USER, &tid, A_mac_categories, &value);
            error_exit(res);
            exit(0);
          }

      default:
        printf("%s (RSBAC %s)\n***\n", argv[0], VERSION);
        printf("Use: %s user attribute [position] value\n\n", argv[0]);  
        printf("- attribute (string) and values = see following list:\n");
        printf("mac_categories\t\t(with additional parameter position)\n\t\t\t0=no, 1=yes\n");
        for (j=0;j<RSBAC_USER_NR_ATTRIBUTES;j++)
          {
            get_attribute_name(tmp1,attr_list[j]);
            get_attribute_param(tmp2,attr_list[j]);
            printf("%s \t%s\n",tmp1,tmp2);
          }
    }
  exit(1);
}
