#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/kdev_t.h>
#include <rsbac/types.h>
#include <rsbac/syscalls.h>
#include <rsbac/getname.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

int main(int argc, char ** argv)
{
  int res = 0;
  int j;
  rsbac_mac_category_vector_t categories;
  char name[80];
  char * progname;
  int verbose = 0;
  
  progname = argv[0];
  if ((argc > 1) && (!strcmp("-v",argv[1])) )
    {
      verbose=1;
      argv++;
      argc--;
    }  
  if (argc >= 3)
    {
      if(strlen(argv[1]) != RSBAC_MAC_NR_CATS)
        {
          fprintf(stderr, "%s: Invalid category string length %i, must be %i\n", progname,
                  strlen(argv[1]),
                  RSBAC_MAC_NR_CATS);
          if(strlen(argv[1]) < RSBAC_MAC_NR_CATS / 2)
            {
              categories = strtol(argv[1],0,0);
              fprintf(stderr, "%s: Using numeric value %lu instead\n", progname,
                      (u_long) categories);
            }
          else
            exit(1);
        }
      else
        {
          for(j=0;j<strlen(argv[1]);j++)
            {
              if(   argv[1][j] != '0'
                 && argv[1][j] != '1')
                {
                  fprintf(stderr, "%s: Invalid attribute value char, must be 0 or 1\n", progname);
                  exit(1);
                }
            }
          strtou64mac(argv[1], &categories);
        }
      if(verbose)
        printf("%s: executing %s with mac_max_categories vector\n%s\n",
               progname,argv[2],u64tostrmac(name, categories));
      res = rsbac_mac_set_max_categories(&categories);
      error_exit(res);
      res = execvp(argv[2],&argv[2]);
      error_exit(res);
    }
  else
    {
      printf("%s (RSBAC %s)\n***\n", progname, VERSION);
      printf("Use: %s [-v] category-vector prog args\n", progname);
      printf("This program will set rsbac_mac_max_categories to category-vector and then\n");
      printf("execute prog with args via execvp()\n");
      printf("-v = verbose\n");
    }
  return (res);
}

