#include <stdio.h>
#include <stdlib.h>
#include <rsbac/types.h>
#include <rsbac/rc_data_structures.h>
#include <rsbac/getname.h>
#include <rsbac/rc_getname.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

int main(int argc, char ** argv)
{
  int res = 0;
  int comp_type;
  enum rsbac_adf_request_t request;
  rsbac_request_vector_t request_vector = 0;
  char tmp1[80],tmp2[120];
  int j;
  enum rsbac_rc_target_t target;
  union rsbac_rc_target_id_t tid;
  enum rsbac_rc_item_t item;
  union rsbac_rc_item_value_t value;
  char * progname = argv[0];
  int role_entry_item_list[RSBAC_RC_NR_ROLE_ENTRY_ITEMS] = RSBAC_RC_ROLE_ENTRY_ITEM_LIST;
  int type_entry_item_list[RSBAC_RC_NR_TYPE_ENTRY_ITEMS] = RSBAC_RC_TYPE_ENTRY_ITEM_LIST;
  int bit_value;
  
  switch(argc)
    {
      case 5:
        target = get_rc_target_nr(argv[1]);
        switch(target)
          {
            case RT_ROLE:
              tid.role = strtol(argv[2],0,10);
              break;
            case RT_TYPE:
              tid.type = strtol(argv[2],0,10);
              break;
            default:
              fprintf(stderr, "Invalid target %s\n", argv[1]);
              exit(1);
           }
        item = get_rc_item_nr(argv[3]);
        if(item == RI_none)
          {
            fprintf(stderr, "Invalid item %s\n", argv[3]);
            exit(1);
          }
        switch (item)
          {
            case RI_role_comp:
              if(strlen(argv[4]) != RSBAC_RC_NR_ROLES)
                {
                  fprintf(stderr, "Invalid bit string len\n");
                  exit(1);
                }
              strtou64rc(argv[4],&value.role_comp);
              break;
            case RI_type_comp_fd:
            case RI_type_comp_dev:
            case RI_type_comp_process:
            case RI_type_comp_ipc:
            case RI_type_comp_scd:
              fprintf(stderr, "parameter comp_type missing\n");
              exit(1);
            case RI_name:
            case RI_type_fd_name:
            case RI_type_dev_name:
            case RI_type_ipc_name:
            case RI_type_process_name:
              if(strlen(argv[4]) >= RSBAC_RC_NAME_LEN)
                {
                  fprintf(stderr, "Name string too long\n");
                  exit(1);
                }
              strcpy(value.name, argv[4]);
              break;
            case RI_admin_type:
              value.admin_type = get_rc_admin_nr(argv[4]);
              if(value.admin_type == RC_none)
                value.admin_type = strtol(argv[4],0,0);
              break;
            default:
              value.dummy = strtol(argv[4],0,0);
          }
        res = rsbac_rc_set_item(target, &tid, item, &value);
        error_exit(res);
        exit(0);
        break;

      case 6:
        target = get_rc_target_nr(argv[1]);
        switch(target)
          {
            case RT_ROLE:
              tid.role = strtol(argv[2],0,10);
              break;
            case RT_TYPE:
              tid.type = strtol(argv[2],0,10);
              break;
            default:
              fprintf(stderr, "Invalid target %s\n", argv[1]);
              exit(1);
           }
        item = get_rc_item_nr(argv[3]);
        if(item == RI_none)
          {
            fprintf(stderr, "Invalid item %s\n", argv[3]);
            exit(1);
          }
        comp_type = strtol(argv[4],0,10);
        if(comp_type > RSBAC_RC_MAX_TYPE)
          {
            fprintf(stderr, "Invalid comp_type %s\n", argv[4]);
            exit(1);
          }
        switch (item)
          {
            case RI_role_comp:
              bit_value = strtol(argv[5],0,0);
              value.dummy = -1;
              res = rsbac_rc_get_item(target, &tid, item, &value);
              error_exit(res);
              if(bit_value)
                value.role_comp |= ((rsbac_rc_type_vector_t) 1 << comp_type);
              else
                value.role_comp &= ~((rsbac_rc_type_vector_t) 1 << comp_type);
              break;
            case RI_type_comp_fd:
            case RI_type_comp_dev:
            case RI_type_comp_process:
            case RI_type_comp_ipc:
            case RI_type_comp_scd:
              if(strlen(argv[5]) != R_NONE)
                {
                  fprintf(stderr, "Invalid bit string len, should be %i\n",
                          R_NONE);
                  exit(1);
                }
              strtou64log(argv[5],&value.type_comp[comp_type]);
              break;
            case RI_name:
            case RI_type_fd_name:
            case RI_type_dev_name:
            case RI_type_ipc_name:
            case RI_type_process_name:
              if(strlen(argv[5]) >= RSBAC_RC_NAME_LEN)
                {
                  fprintf(stderr, "Name string too long\n");
                  exit(1);
                }
              strcpy(value.name, argv[5]);
              break;
            case RI_admin_type:
              value.admin_type = get_rc_admin_nr(argv[5]);
              if(value.admin_type == RC_none)
                value.admin_type = strtol(argv[5],0,0);
              break;
            default:
              value.dummy = strtol(argv[5],0,0);
          }
        res = rsbac_rc_set_item(target, &tid, item, &value);
        error_exit(res);
        exit(0);
        break;

      case 7:
        target = get_rc_target_nr(argv[1]);
        switch(target)
          {
            case RT_ROLE:
              tid.role = strtol(argv[2],0,10);
              break;
            case RT_TYPE:
              tid.type = strtol(argv[2],0,10);
              break;
            default:
              fprintf(stderr, "Invalid target %s\n", argv[1]);
              exit(1);
           }
        item = get_rc_item_nr(argv[3]);
        switch(item)
          {
            case RI_none:
              fprintf(stderr, "Invalid item %s\n", argv[3]);
              exit(1);
            case RI_type_comp_fd:
              request_vector = RSBAC_FD_REQUEST_VECTOR;
              break;
            case RI_type_comp_dev:
              request_vector = RSBAC_DEV_REQUEST_VECTOR;
              break;
            case RI_type_comp_process:
              request_vector = RSBAC_PROCESS_REQUEST_VECTOR;
              break;
            case RI_type_comp_ipc:
              request_vector = RSBAC_IPC_REQUEST_VECTOR;
              break;
            case RI_type_comp_scd:
              request_vector = RSBAC_SCD_REQUEST_VECTOR | RSBAC_SYSTEM_REQUEST_VECTOR;
              break;
            default:
              request_vector = RSBAC_ALL_REQUEST_VECTOR;
          }
        comp_type = strtol(argv[4],0,10);
        if(comp_type > RSBAC_RC_MAX_TYPE)
          {
            fprintf(stderr, "Invalid comp_type %s\n", argv[4]);
            exit(1);
          }
        request = get_request_nr(argv[5]);
        if(request == R_NONE)
          {
            request = strtol(argv[5],0,10);
            if(   (request >= R_NONE)
               || (   (request == 0)
                   && strcmp(argv[5],"0")
                  )
              )
              {
                if(!strcmp(argv[5],"R"))
                  {
                    request_vector &= RSBAC_READ_REQUEST_VECTOR;
                  }
                else
                if(!strcmp(argv[5],"RW"))
                  {
                    request_vector &= RSBAC_READ_WRITE_REQUEST_VECTOR;
                  }
                else
                if(!strcmp(argv[5],"SY"))
                  {
                    request_vector &= RSBAC_SYSTEM_REQUEST_VECTOR;
                  }
                else
                if(!strcmp(argv[5],"SE"))
                  {
                    request_vector &= RSBAC_SECURITY_REQUEST_VECTOR;
                  }
                else
                if(!strcmp(argv[5],"A"))
                  {
                    request_vector &= RSBAC_ALL_REQUEST_VECTOR;
                  }
                else
                  {
                    fprintf(stderr, "Invalid request %s\n", argv[4]);
                    exit(1);
                  }
              }
          }
        else
          {
            request_vector &= ((rsbac_rc_request_vector_t) 1 << request);
          }
        bit_value = strtol(argv[6],0,0);
        switch (item)
          {
            case RI_role_comp:
              value.dummy = -1;
              res = rsbac_rc_get_item(target, &tid, item, &value);
              error_exit(res);
              if(bit_value)
                value.role_comp |= ((rsbac_rc_type_vector_t) 1 << comp_type);
              else
                value.role_comp &= ~((rsbac_rc_type_vector_t) 1 << comp_type);
              break;
            case RI_type_comp_fd:
            case RI_type_comp_dev:
            case RI_type_comp_process:
            case RI_type_comp_ipc:
            case RI_type_comp_scd:
              value.dummy = -1;
              res = rsbac_rc_get_item(target, &tid, item, &value);
              error_exit(res);
              if(bit_value)
                value.type_comp[comp_type] |= request_vector;
              else
                value.type_comp[comp_type] &= ~request_vector;
              break;
            case RI_name:
            case RI_type_fd_name:
            case RI_type_dev_name:
            case RI_type_ipc_name:
            case RI_type_process_name:
              if(strlen(argv[6]) >= RSBAC_RC_NAME_LEN)
                {
                  fprintf(stderr, "Name string too long\n");
                  exit(1);
                }
              strcpy(value.name, argv[6]);
              break;
            case RI_admin_type:
              value.admin_type = get_rc_admin_nr(argv[6]);
              if(value.admin_type == RC_none)
                value.admin_type = strtol(argv[6],0,0);
              break;
            default:
              value.dummy = strtol(argv[6],0,0);
          }
        res = rsbac_rc_set_item(target, &tid, item, &value);
        error_exit(res);
        exit(0);
        break;

      default:
        printf("%s (RSBAC %s)\n***\n", progname, VERSION);
        printf("Use: %s rc-target-type id-nr item [role/type [request]] value\n", progname);  
        printf("- rc-target-type = ROLE or TYPE,\n");
        printf("- id-nr = ROLE or TYPE number,\n");
        printf("- item = entry line,\n");
        printf("- role/type = for this type only (required for type_comp items only),\n");
        printf("- request = request name or number (type_comp items only),\n");
        printf("        also special request groups R (read requests),\n");
        printf("        RW (read-write), SY (system), SE (security), A (all)\n");
        printf("- item (string) and value = see following list:\n");
        printf("- ROLE:\n");
        for (j=0;j<RSBAC_RC_NR_ROLE_ENTRY_ITEMS;j++)
          {
            get_rc_item_name(tmp1,role_entry_item_list[j]);
            get_rc_item_param(tmp2,role_entry_item_list[j]);
            printf("%s \t%s\n",tmp1,tmp2);
          }
        printf("- TYPE:\n");
        for (j=0;j<RSBAC_RC_NR_TYPE_ENTRY_ITEMS;j++)
          {
            get_rc_item_name(tmp1,type_entry_item_list[j]);
            get_rc_item_param(tmp2,type_entry_item_list[j]);
            printf("%s \t%s\n",tmp1,tmp2);
          }
    }
  exit(1);
}
